@@ -3,15 +3,16 @@
  *	Original interface by Otto Lind, 6/3/93
  *	Additional map and map! support by Kevin Buettner, 9/17/94
  *
- * $Header: /usr2/foxharp/src/pgf/vile/RCS/map.c,v 1.53 1995/02/06 04:06:39 pgf Exp $
- * 
+ * $Header: /usr/build/vile/vile/RCS/map.c,v 1.97 2002/11/01 23:58:39 tom Exp $
+ *
  */
 
 #include "estruct.h"
 #include "edef.h"
 
-#define MAXLHS	NSTRING
-
+#if DISP_NTWIN
+# define NTWIN_CURSOR_MINWAIT 550   /* msec */
+#endif
 
 /*
  * Picture for struct maprec
@@ -51,17 +52,17 @@
 
 struct maprec {
 	int		ch;		/* character to match		*/
-	int		flags;		/* flags word			*/
+	UINT		flags;		/* flags word			*/
 	struct maprec *	dlink;		/* Where to go to match the	*/
 					/*   next character in a multi-	*/
 					/*   character sequence.	*/
 	struct maprec *	flink;		/* Where to try next if match	*/
-					/*   against current character 	*/
+					/*   against current character	*/
 					/*   is unsuccessful		*/
 	int		irv;		/* system defined mapping: The	*/
 					/*   (wide) character code to	*/
 					/*   replace a matched sequence by */
-	char          *	srv;		/* user defined mapping: This	*/
+	char	      * srv;		/* user defined mapping: This	*/
 					/*   is the string to replace a	*/
 					/*   matched sequence by	*/
 };
@@ -78,60 +79,58 @@ static struct maprec *map_insert = NULL;
 static struct maprec *map_syskey = NULL;
 static struct maprec *abbr_map = NULL;
 
-static	int	map_common P(( struct maprec **, char *, int ));
-static	int	unmap_common P(( struct maprec **, char * ));
-static	void	addtomap P(( struct maprec **, char *, int, int, int, char * ));
-static	int	delfrommap P(( struct maprec **, char * ));
+static	int	map_common(struct maprec **mpp, const char *bufname, UINT remapflag);
+static	int	unmap_common(struct maprec **mpp, const char *bufname);
+static	void	addtomap(struct maprec **mpp, const char * ks, int kslen, UINT flags, int irv, char * srv);
+static	int	delfrommap(struct maprec **mpp, const char * ks);
 
-static	int	abbr_getc P(( void ));
-static	int	abbr_c_avail P(( void ));
+static	int	abbr_getc (void);
+static	int	abbr_c_avail (void);
 
-static int	mapgetc P(( void ));
+static	int	mapgetc (void);
 
-typedef	int (*AvailFunc) P(( void ));
-typedef	int (*GetFunc) P(( void ));
+typedef	int (*AvailFunc) (void);
+typedef	int (*GetFunc) (void);
+typedef	int (*StartFunc) (void);
 
-static int maplookup P(( int, ITBUFF **, struct maprec *, GetFunc, AvailFunc ));
+static	int	maplookup(int c, ITBUFF **outp, struct maprec *mp, GetFunc get, AvailFunc avail, StartFunc start, int suffix);
 
-#if OPT_SHOW_MAPS
-static void makemaplist P(( int, void * ));
-static int show_mapped_chars P(( char * ));
-#endif
-
-#if OPT_SHOW_MAPS && OPT_UPBUFF
-static int show_Mappings P(( BUFFER * ));
-#endif
-
-#if OPT_UPBUFF
-static void relist_mappings P(( char * ));
-#else
+#if !OPT_UPBUFF
 #define relist_mappings(name)
 #endif
 
+#define NUMKEYSTR (KBLOCK / 2)
+
+#ifdef DOKEYLOG
+int do_keylog = 1;
+#endif
+
 #if OPT_SHOW_MAPS
 #define MAPS_PREFIX 12
 
 /*ARGSUSED*/
 static void
-makemaplist(dummy, mapp)
-    int   dummy;
-    void *mapp;
+makemaplist(int dummy GCC_UNUSED, void *mapp)
 {
-    char lhsstr[MAXLHS];
-    struct maprec *lhsstack[MAXLHS];
+    TBUFF *lhsstr = 0;
+    struct maprec **lhsstack;
     struct maprec *mp = (struct maprec *) mapp;
-    int depth = 0;
     int footnote = 0;
-    int i;
+    size_t depth = 0;
+    size_t maxdepth;
+    size_t i;
 
-    for (;;) {
+    lhsstr = tb_init(&lhsstr, 0);
+    lhsstack = typeallocn(struct maprec *, maxdepth = NSTRING);
+    for_ever {
 	if (mp) {
-	    char *remapnote;
+	    const char *remapnote;
 	    char *mapstr;
-	    char esc_seq[10];
-	    lhsstr[depth] = mp->ch;
+	    char esc_seq[10];	/* FIXME */
+	    tb_put(&lhsstr, depth, mp->ch);
+	    if (depth+1 >= maxdepth)
+		lhsstack = typereallocn(struct maprec *, lhsstack, maxdepth *= 2);
 	    lhsstack[depth++] = mp->flink;
-	    lhsstr[depth] = 0;
 
 	    mapstr = (char *)0;
 	    if (mp->srv) {
@@ -141,10 +140,10 @@ makemaplist(dummy, mapp)
 		mapstr = esc_seq;
 	    }
 	    if (mapstr) {
-    		    if (mapp && (struct maprec *)mapp == abbr_map) {
+		    if (mapp && (struct maprec *)mapp == abbr_map) {
 			    /* the abbr map is stored inverted */
-			    for (i = depth; i > 0; )
-				bputc(lhsstr[--i]);	/* may contain nulls */
+			    for (i = depth; i != 0; )
+				bputc(tb_values(lhsstr)[--i]);
 		    } else {
 			    if (mp->flags & MAPF_NOREMAP) {
 				remapnote = "(n)";
@@ -154,13 +153,13 @@ makemaplist(dummy, mapp)
 			    }
 			    bprintf("%s ", remapnote);
 			    for (i = 0; i < depth; i++)
-				bputc(lhsstr[i]);	/* may contain nulls */
+				bputc(tb_values(lhsstr)[i]);
 		    }
 		    bprintf("\t%s\n", mapstr);
 	    }
 	    mp = mp->dlink;
 	}
-	else if (depth > 0)
+	else if (depth != 0)
 	    mp = lhsstack[--depth];
 	else
 	    break;
@@ -168,11 +167,12 @@ makemaplist(dummy, mapp)
     if (footnote) {
 	bprintf("[(n) means never remap]\n");
     }
+    tb_free(&lhsstr);
+    free((char *)lhsstack);
 }
 
 static int
-show_mapped_chars(bname)
-    char *bname;
+show_mapped_chars(const char *bname)
 {
 	struct maprec *mp;
 	if (strcmp(bname, MAP_BufName) == 0)
@@ -190,8 +190,7 @@ show_mapped_chars(bname)
 
 #if OPT_UPBUFF
 static int
-show_Mappings(bp)
-BUFFER *bp;
+show_Mappings(BUFFER *bp)
 {
 	b_clr_obsolete(bp);
 	return show_mapped_chars(bp->b_bname);
@@ -200,8 +199,7 @@ BUFFER *bp;
 #undef relist_mappings
 
 static void
-relist_mappings(bufname)
-    char * bufname;
+relist_mappings(const char * bufname)
 {
     update_scratch(bufname, show_Mappings);
 }
@@ -214,40 +212,35 @@ relist_mappings(bufname)
 */
 /* ARGSUSED */
 int
-map(f, n)
-    int f, n;
+map(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return map_common(&map_command, MAP_BufName, 0);
 }
 
 /* ARGSUSED */
 int
-map_bang(f, n)
-    int f, n;
+map_bang(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return map_common(&map_insert, MAPBANG_BufName, 0);
 }
 
 /* ARGSUSED */
 int
-noremap(f, n)
-    int f, n;
+noremap(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return map_common(&map_command, MAP_BufName, MAPF_NOREMAP);
 }
 
 /* ARGSUSED */
 int
-noremap_bang(f, n)
-    int f, n;
+noremap_bang(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return map_common(&map_insert, MAPBANG_BufName, MAPF_NOREMAP);
 }
 
 /* ARGSUSED */
 int
-abbrev(f, n)
-    int f, n;
+abbrev(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return map_common(&abbr_map, ABBR_BufName, MAPF_NOREMAP);
 }
@@ -255,22 +248,18 @@ abbrev(f, n)
 #if OPT_SHOW_MAPS
 /* ARGSUSED */
 int
-sysmap(f, n)
-int f, n;
+sysmap(int f GCC_UNUSED, int n GCC_UNUSED)
 {
 	return show_mapped_chars(SYSMAP_BufName);
 }
 #endif
 
 static int
-map_common(mpp, bufname, remapflag)
-    struct maprec **mpp;
-    char *bufname;
-    int remapflag;
+map_common(struct maprec **mpp, const char *bufname, UINT remapflag)
 {
     int	 status;
-    char kbuf[NSTRING];
-    char val[NSTRING];
+    static TBUFF *kbuf;
+    static TBUFF *val;
     int len;
 
 #if OPT_SHOW_MAPS
@@ -278,37 +267,38 @@ map_common(mpp, bufname, remapflag)
 	return show_mapped_chars(bufname);
     }
 #endif
-    kbuf[0] = EOS;
-    status = kbd_string("change this string: ", kbuf, sizeof(kbuf),
+    tb_scopy(&kbuf, "");
+    status = kbd_reply("change this string: ", &kbuf, eol_history,
 			' ', KBD_NOMAP|KBD_NOEVAL, no_completion);
     if (status != TRUE)
 	return status;
 
     hst_glue(' ');
-    val[0] = EOS;
+    tb_scopy(&val, "");
     if (!clexec) {
-	    status = kbd_string("to this new string: ", val, sizeof(val),
+	    status = kbd_reply("to this new string: ", &val, eol_history,
 			'\n', KBD_NOMAP, no_completion);
     } else {
-	    (void)macliteralarg(val); /* consume to end of line */
-	    status = (*val != EOS);
+	    (void)mac_literalarg(&val); /* consume to end of line */
+	    status = tb_length(val) > 1;
     }
     if (status != TRUE)
 	return status;
 
-
-    len = strlen(kbuf);
+    len = tb_length(kbuf) - 1;
     if ((*mpp && *mpp == abbr_map) || (strcmp(bufname, ABBR_BufName) == 0)) {
 	/* reverse the lhs */
-	int i, t;
+	int i;
+	char t;
+	char *s = tb_values(kbuf);
 	for (i = 0; i < len/2; i++) {
-	    t = kbuf[len-i-1];
-	    kbuf[len-i-1] = kbuf[i];
-	    kbuf[i] = t;
+	    t = s[len-i-1];
+	    s[len-i-1] = s[i];
+	    s[i] = t;
 	}
     }
 
-    addtomap(mpp, kbuf, len, MAPF_USERTIMER|remapflag, -1, val);
+    addtomap(mpp, tb_values(kbuf), len, MAPF_USERTIMER|remapflag, -1, tb_values(val));
     relist_mappings(bufname);
     return TRUE;
 }
@@ -318,50 +308,44 @@ map_common(mpp, bufname, remapflag)
 */
 /* ARGSUSED */
 int
-unmap(f, n)
-    int f,n;
+unmap(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return unmap_common(&map_command, MAP_BufName);
 }
 
 /* ARGSUSED */
 int
-unmap_bang(f, n)
-    int f, n;
+unmap_bang(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return unmap_common(&map_insert, MAPBANG_BufName);
 }
 
 /* ARGSUSED */
 int
-unmap_system(f, n)
-    int f, n;
+unmap_system(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return unmap_common(&map_syskey, SYSMAP_BufName);
 }
 
 /* ARGSUSED */
 int
-unabbr(f, n)
-    int f, n;
+unabbr(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     return unmap_common(&abbr_map, ABBR_BufName);
 }
 
 static int
-unmap_common(mpp, bufname)
-    struct maprec **mpp;
-    char *bufname;
+unmap_common(struct maprec **mpp, const char *bufname)
 {
     int	 status;
-    char kbuf[NSTRING];
+    static TBUFF *kbuf;
 
     /* otherwise it'll be mapped, and not found when looked up */
     if (mpp && mpp == &map_syskey)
-    	suppress_sysmap = TRUE;
+	suppress_sysmap = TRUE;
 
-    kbuf[0] = EOS;
-    status = kbd_string("unmap string: ", kbuf, sizeof(kbuf),
+    tb_scopy(&kbuf, "");
+    status = kbd_reply("unmap string: ", &kbuf, eol_history,
 			' ', KBD_NOMAP, no_completion);
     suppress_sysmap = FALSE;
     if (status != TRUE)
@@ -369,43 +353,42 @@ unmap_common(mpp, bufname)
 
     if ((*mpp && *mpp == abbr_map) || (strcmp(bufname, ABBR_BufName) == 0)) {
 	/* reverse the lhs */
-	int i, t;
-    	int len = strlen(kbuf);
+	int i;
+	char t;
+	int len = tb_length(kbuf) - 1;
+	char *s = tb_values(kbuf);
 	for (i = 0; i < len/2; i++) {
-	    t = kbuf[len-i-1];
-	    kbuf[len-i-1] = kbuf[i];
-	    kbuf[i] = t;
+	    t = s[len-i-1];
+	    s[len-i-1] = s[i];
+	    s[i] = t;
 	}
     }
 
-    if (delfrommap(mpp, kbuf) != TRUE) {
+    if (delfrommap(mpp, tb_values(kbuf)) != TRUE) {
 	mlforce("[Sequence not mapped]");
 	return FALSE;
     }
     relist_mappings(bufname);
     return TRUE;
 }
-    
+
 /* addtosysmap is used to initialize the system default function key map
 */
 void
-addtosysmap(seq, seqlen, code)
-    char * seq;
-    int    seqlen;
-    int    code;
+addtosysmap(const char * seq, int seqlen, int code)
 {
     addtomap(&map_syskey, seq, seqlen, MAPF_SYSTIMER,
-    			code, (char *)0);
+			code, (char *)0);
 }
 
 static void
-addtomap(mpp, ks, kslen, flags, irv, srv)
-    struct maprec **mpp;
-    char *	ks;
-    int         kslen;
-    int         flags;
-    int		irv;
-    char *	srv;
+addtomap(
+    struct maprec **mpp,
+    const char * ks,
+    int 	kslen,
+    UINT	flags,
+    int		irv,
+    char *	srv)
 {
     struct maprec *mp = NULL;
 
@@ -415,7 +398,7 @@ addtomap(mpp, ks, kslen, flags, irv, srv)
     while (*mpp && kslen) {
 	mp = *mpp;
 	mp->flags |= flags;
-	if (*ks == mp->ch) {
+	if (char2int(*ks) == mp->ch) {
 	    mpp = &mp->dlink;
 	    ks++;
 	    kslen--;
@@ -430,7 +413,7 @@ addtomap(mpp, ks, kslen, flags, irv, srv)
 	    break;
 	*mpp = mp;
 	mp->dlink = mp->flink = NULL;
-	mp->ch = *ks++;
+	mp->ch = char2int(*ks++);
 	mp->srv = NULL;
 	mp->flags = flags;
 	mp->irv = -1;
@@ -449,29 +432,38 @@ addtomap(mpp, ks, kslen, flags, irv, srv)
 }
 
 static int
-delfrommap(mpp, ks)
-    struct maprec **mpp;
-    char * ks;
+delfrommap(struct maprec **mpp, const char * ks)
 {
-    struct maprec **mstk[MAXLHS];
+    struct maprec **save_m = mpp;
+    struct maprec ***mstk = 0;
+    const char *save_k = ks;
     int depth = 0;
+    int pass;
 
     if (ks == 0 || *ks == 0)
 	return FALSE;
 
-    while (*mpp && *ks) {
-	mstk[depth] = mpp;
-	if ((*mpp)->ch == *ks) {
-	    mpp = &(*mpp)->dlink;
-	    ks++;
-	    depth++;
+    for (pass = 0; pass < 2; pass++) {
+	mpp   = save_m;
+	ks    = save_k;
+	depth = 0;
+	while (*mpp && *ks) {
+	    if (pass)
+		mstk[depth] = mpp;
+	    if ((*mpp)->ch == char2int(*ks)) {
+		mpp = &(*mpp)->dlink;
+		ks++;
+		depth++;
+	    }
+	    else
+		mpp = &(*mpp)->flink;
 	}
-	else
-	    mpp = &(*mpp)->flink;
-    }
 
-    if (*ks)
-	return FALSE;		/* not in map */
+	if (*ks)
+	    return FALSE;		/* not in map */
+	if (!pass)
+	    mstk = typecallocn(struct maprec **, depth+1);
+    }
 
     depth--;
     if ((*mstk[depth])->srv) {
@@ -480,6 +472,7 @@ delfrommap(mpp, ks)
     } else if ((*mstk[depth])->irv != -1) {
 	(*mstk[depth])->irv = -1;
     } else {
+	free((char *)mstk);
 	return FALSE;
     }
 
@@ -494,6 +487,7 @@ delfrommap(mpp, ks)
 	else
 	    break;
     }
+    free((char *)mstk);
     return TRUE;
 }
 
@@ -502,25 +496,76 @@ delfrommap(mpp, ks)
 #define INPUT_FROM_MAPGETC 2
 
 static ITBUFF *sysmappedchars = NULL;
+static ITBUFF *mapgetc_ungottenchars = NULL;
+static int mapgetc_ungotcnt = 0;
+
+static void
+save_keystroke(int c)
+{
+    KILL *kp;
+    KILLREG *kr = &kbs[KEYST_KREG];
+
+    beginDisplay();
+#ifdef DOKEYLOG
+    if (do_keylog) {
+	static int keyfd = -1;
+	static char *tfilenam;
+	if (!tfilenam)
+	    tfilenam = tempnam("/tmp/vilekeylogs", "vilek");
+	if (tfilenam) {
+	    if (keyfd < 0)
+		keyfd = open(tfilenam, O_CREAT | O_WRONLY, 0600);
+	    if (keyfd >= 0)
+		write(keyfd, &c, 1);
+	}
+    }
+#endif
+    if (kr->kbufh == NULL) {
+	kr->kbufh = typealloc(KILL);
+	kr->kused = 0;
+    }
+    if (kr->kbufh != NULL) {
+
+	kp = kr->kbufp = kr->kbufh;
+	kp->d_next = NULL;
+
+	kp->d_chunk[kr->kused++] = (UCHAR) c;
+	if (kr->kused >= NUMKEYSTR * 2) {	/* time to dump the oldest half */
+	    (void) memcpy((kp->d_chunk),
+			  (&kp->d_chunk[NUMKEYSTR / 2]),
+			  NUMKEYSTR / 2);
+	    kr->kused = NUMKEYSTR / 2;
+	}
+    }
+    endofDisplay();
+}
 
 /* these two wrappers are provided because at least one pcc-based
-	compiler balks at passing TTgetc or TTtypahead as a function pointer */
-static int normal_getc P(( void ));
-static int normal_typeahead P(( void ));
+	compiler balks at passing term.getch or term.typahead as a function pointer */
 
 static int
-normal_getc()
+normal_getc(void)
 {
-      return(TTgetc());
+	int c = term.getch();
+	TRACE(("normal/getc:%c (%#x)\n", c, c));
+	save_keystroke(c);
+	return c;
 }
+
 static int
-normal_typeahead()
+normal_typeahead(void)
 {
-      return(TTtypahead());
+	return(term.typahead());
+}
+
+static int
+normal_start(void)
+{
+	return TRUE;
 }
 
 int
-sysmapped_c()
+sysmapped_c(void)
 {
     int c;
 
@@ -528,57 +573,69 @@ sysmapped_c()
     if (itb_more(sysmappedchars))
 	return itb_last(sysmappedchars);
 
-    c = TTgetc();
+    if ((c = term.getch()) == -1)
+	return c;	/* see comment in ttgetc */
+
+    TRACE(("mapped/getc:%c (%#x)\n", c, c));
+
+    save_keystroke(c);
 
     if (suppress_sysmap)
-    	return c;
+	return c;
 
     /* will push back on sysmappedchars successful, or not */
-    (void)maplookup(c, &sysmappedchars, map_syskey, 
-    		normal_getc, normal_typeahead);
+    (void)maplookup(c, &sysmappedchars, map_syskey,
+		normal_getc, normal_typeahead, normal_start, TRUE);
 
     return itb_last(sysmappedchars);
 }
 
 int
-sysmapped_c_avail()
+sysmapped_c_avail(void)
 {
-    return itb_more(sysmappedchars) || TTtypahead();
+    return itb_more(sysmappedchars) || (!mapgetc_ungotcnt && term.typahead());
 }
 
 
-static ITBUFF *mapgetc_ungottenchars = NULL;
-static int mapgetc_ungotcnt;
+/* make the assumption that no input will magically appear
+ * (un)available to tgetc in between a mapungetc and the next mapgetc.
+ * Hence characters can't be ungotten onto the wrong buffer (exception
+ * is the last tgot char might be mapungot onto the map buffer.  This
+ * is OK (if assumption holds) because the next character will be
+ * gotten from this buffer.
+ */
 
 void
-mapungetc(c)
-int c;
+mapungetc(int c)
 {
-	itb_append(&mapgetc_ungottenchars, c);
+    if (tgetc_avail()) {
+	tungetc(c);
+    } else {
+	(void)itb_append(&mapgetc_ungottenchars, c);
 	mapgetc_ungotcnt++;
+    }
 }
 
-#define TOOMANY 1200
 static int infloopcount;
 static int mapgetc_raw_flag;
 
 static int
-mapgetc()
+mapgetc(void)
 {
-    int remapflag;
+    UINT remapflag;
     if (global_g_val(GMDREMAP))
-    	remapflag = 0;
+	remapflag = 0;
     else
-    	remapflag = NOREMAP;
+	remapflag = NOREMAP;
 
-    if (mapgetc_ungotcnt > 0) {
-	    if (infloopcount++ > TOOMANY) {
-		itb_init(&mapgetc_ungottenchars, abortc);
+    if (!tgetc_avail() && mapgetc_ungotcnt > 0) {
+	    if (infloopcount++ > global_g_val(GVAL_MAPLENGTH)) {
+		(void)itb_init(&mapgetc_ungottenchars, esc_c);
 		mapgetc_ungotcnt = 0;
 		mlforce("[Infinite loop detected in %s sequence]",
 			    (insertmode) ? "map!" : "map");
-		catnap(1000,FALSE);  /* FIXX: be sure message gets out */
-		return abortc|NOREMAP;
+		catnap(1000,FALSE);  /* FIXME: be sure message gets out */
+		return esc_c|NOREMAP;
 	    }
 	    mapgetc_ungotcnt--;
 	    return itb_last(mapgetc_ungottenchars) | remapflag;
@@ -588,48 +645,64 @@ mapgetc()
 }
 
 int
-mapped_c_avail()
+mapped_c_avail(void)
 {
     return mapgetc_ungotcnt > 0 || tgetc_avail();
 }
 
 int
-mapped_c(remap, raw)
-int remap;
-int raw;
+mapped_ungotc_avail(void)
+{
+    return mapgetc_ungotcnt > 0;
+}
+
+static int
+mapped_c_start(void)
+{
+    return TRUE;
+}
+
+int
+mapped_c(int remap, int raw)
 {
     int c;
     int matched;
     struct maprec *mp;
     int speckey = FALSE;
     static ITBUFF *mappedchars = NULL;
-    
+
     /* still some pushback left? */
     mapgetc_raw_flag = raw;
     c = mapgetc();
 
-    if (!remap || (c & NOREMAP))
-	return (c & ~NOREMAP);
+    if ((c & YESREMAP) == 0 && (!remap || (c & NOREMAP)))
+	return (c & ~REMAPFLAGS);
+
+    c &= ~REMAPFLAGS;
 
     if (reading_msg_line)
-    	mp = 0;
+	mp = 0;
     else if (insertmode)
-    	mp = map_insert;
-    else 
-    	mp = map_command;
+	mp = map_insert;
+    else
+	mp = map_command;
 
     /* if we got a function key from the lower layers, turn it into '#c'
-    	and see if the user remapped that */
-    if (c & SPEC) {
+	and see if the user remapped that */
+    if ((c & SPEC)
+#if OPT_KEY_MODIFY
+     && !(c & mod_KEY)	/* we don't do this special case */
+#endif
+     ) {
 	mapungetc(kcod2key(c));
 	c = poundc;
 	speckey = TRUE;
     }
 
     do {
-	itb_init(&mappedchars, abortc);
+	(void)itb_init(&mappedchars, esc_c);
 
-	matched = maplookup(c, &mappedchars, mp, mapgetc, mapped_c_avail);
+	matched = maplookup(c, &mappedchars, mp, mapgetc, mapped_c_avail, mapped_c_start, TRUE);
 
 
 	while(itb_more(mappedchars))
@@ -639,10 +712,10 @@ int raw;
 	    in the first place.  unless they wanted it quoted.  then we
 	    leave it as is */
 	if (!raw && speckey && !matched) {
-	    c = mapgetc() & ~NOREMAP;
+	    c = mapgetc() & ~REMAPFLAGS;
 	    if (c != poundc)
 		    dbgwrite("BUG: # problem in mapped_c");
-	    return (mapgetc() & ~NOREMAP) | SPEC;
+	    return (mapgetc() & ~REMAPFLAGS) | SPEC;
 	}
 
 	c = mapgetc();
@@ -652,9 +725,10 @@ int raw;
 
 	speckey = FALSE;
 
-    } while (matched && remap && !(c & NOREMAP) );
+    } while (matched &&
+	((remap && !(c & NOREMAP)) || (c & YESREMAP)) );
 
-    return c & ~NOREMAP;
+    return c & ~REMAPFLAGS;
 
 }
 
@@ -662,58 +736,69 @@ static int abbr_curr_off;
 static int abbr_search_lim;
 
 static int
-abbr_getc()
+abbr_getc(void)
 {
     if (abbr_curr_off <= abbr_search_lim)
-    	return -1; /* won't match anything in the tree */
-    return lGetc(DOT.l, --abbr_curr_off);
+	return -1; /* won't match anything in the tree */
+    return lgetc(DOT.l, --abbr_curr_off) & 0xff;
 }
 
 static int
-abbr_c_avail()
+abbr_c_avail(void)
 {
     return TRUE;
 }
 
+static int
+abbr_c_start(void)
+{
+    if (abbr_curr_off > abbr_search_lim) {
+	/* we need to check the char in front of the match.
+	   if it's a similar type to the first char of the match,
+	   i.e. both idents, or both non-idents, we do nothing.
+	   if whitespace precedes either ident or non-ident, the
+	   match is good.
+	 */
+	int first, prev;
+	first = lgetc(DOT.l,abbr_curr_off);
+	prev = lgetc(DOT.l,abbr_curr_off-1);
+	if ((isident(first) && isident(prev)) ||
+	    (!isident(first) && !(isident(prev) || isSpace(prev)))) {
+		return FALSE;
+	}
+    }
+    return TRUE;
+}
+
 void
-abbr_check(backsp_limit_p)
-int *backsp_limit_p;
+abbr_check(int *backsp_limit_p)
 {
     int matched;
     ITBUFF *abbr_chars = NULL;
     int status = TRUE;
 
-    if (lLength(DOT.l) < 1)
-    	return;
+    if (llength(DOT.l) < 1)
+	return;
     abbr_curr_off = DOT.o;
     abbr_search_lim = *backsp_limit_p;
-    itb_init(&abbr_chars, abortc);
+    (void)itb_init(&abbr_chars, esc_c);
     matched = maplookup(abbr_getc(), &abbr_chars, abbr_map,
-    	abbr_getc, abbr_c_avail);
+	abbr_getc, abbr_c_avail, abbr_c_start, FALSE);
 
 
     if (matched) {
 	    /* there are still some conditions that have to be met by
 	       the preceding chars, if any */
-	    if (abbr_curr_off > abbr_search_lim) {
-		/* we need to check the char in front of the match.
-		   if it's a similar type to the first char of the match, 
-		   i.e. both idents, or both non-idents, we do nothing.
-		   if whitespace precedes either ident or non-ident, the
-		   match is good.
-		 */
-		char first, prev;
-		first = lGetc(DOT.l,abbr_curr_off);
-		prev = lGetc(DOT.l,abbr_curr_off-1);
-		if ((isident(first) && isident(prev)) ||
-		    (!isident(first) && !(isident(prev) || isspace(prev))))
-		    	return;
+	    if (!abbr_c_start()) {
+		itb_free(&abbr_chars);
+		return;
 	    }
 	    DOT.o -= matched;
-	    ldelete(matched, FALSE);
+	    ldelete((B_COUNT)matched, FALSE);
 	    while(status && itb_more(abbr_chars))
 		status = inschar(itb_last(abbr_chars), backsp_limit_p);
     }
+    itb_free(&abbr_chars);
     return;
 }
 
@@ -726,20 +811,23 @@ int *backsp_limit_p;
   returns number of characters matched
 */
 static int
-maplookup(c, outp, mp, get, avail)
-    int c;
-    ITBUFF **outp;
-    struct maprec *mp;
-    GetFunc get;
-    AvailFunc avail;
+maplookup(
+    int c,
+    ITBUFF **outp,
+    struct maprec *mp,
+    GetFunc get,
+    AvailFunc avail,
+    StartFunc start,
+    int suffix)
 {
     struct maprec *rmp = NULL;
-    int unmatched[MAXLHS];
+    ITBUFF *unmatched = 0;
     int matchedcnt;
     int use_sys_timing;
+    int had_start = FALSE;
     register int count = 0;	/* index into 'unmatched[]' */
 
-    /* 
+    /*
      * we don't want to delay for a user-specified :map!  starting with
      * poundc since it's likely that the mapping is happening on behalf of
      * a function key.  (it's so the user can ":map! #1 foobar" but still be
@@ -748,23 +836,31 @@ maplookup(c, outp, mp, get, avail)
      * on it.
      */
     use_sys_timing = (insertmode && c == poundc &&
-    				(isprint(poundc) || isspace(poundc)));
+				(isPrint(poundc) || isSpace(poundc)));
+
+    unmatched = itb_init(&unmatched, 0);
+    itb_append(&unmatched, c);
+    count++;
 
-    unmatched[count++] = c;
     matchedcnt = 0;
     while (mp != 0) {
 	if (c == mp->ch) {
 	    if (mp->irv != -1 || mp->srv != NULL) {
 		rmp = mp;
 		matchedcnt += count;
+		unmatched = itb_init(&unmatched, 0);
 		count = 0;
 
 		/* our code supports matching the longer of two maps one of
 		 * which is a subset of the other.  vi matches the shorter
 		 * one.
 		 */
-	        if (!global_g_val(GMDMAPLONGER))
-		    break;
+		if ((*start)()) {
+		    had_start = TRUE;
+		    if (!global_g_val(GMDMAPLONGER)) {
+			break;
+		    }
+		}
 	    }
 
 	    mp = mp->dlink;
@@ -777,6 +873,9 @@ maplookup(c, outp, mp, get, avail)
 
 		/* give it a little extra time... */
 		int timer = 0;
+#if DISP_NTWIN
+		int cursor_state = 0;
+#endif
 
 		/* we want to use the longer of the two timers */
 
@@ -791,55 +890,75 @@ maplookup(c, outp, mp, get, avail)
 				timer = global_g_val(GVAL_TIMEOUTVAL);
 		}
 
+#if DISP_NTWIN
+		if (timer > NTWIN_CURSOR_MINWAIT) {
+			/*
+			 * editor waits for a potentially long time
+			 * for the next character, during which the
+			 * cursor will be invisible; so turn it back on.
+			 * unfortunately, the cursor won't blink
+			 * because there's no thread driving the
+			 * message pump.
+			 */
+
+			cursor_state = winvile_cursor_state(TRUE, FALSE);
+		}
+#endif
 		catnap(timer,TRUE);
+#if DISP_NTWIN
+		if (timer > NTWIN_CURSOR_MINWAIT) /* restore cursor state */
+			(void) winvile_cursor_state(cursor_state, FALSE);
+#endif
 
 		if (!(*avail)())
 		    break;
 	    }
 
-	    unmatched[count++] = c = (*get)() & ~NOREMAP;
+	    if ((c = (*get)()) < 0)
+		break;
+
+	    itb_append(&unmatched, c & ~REMAPFLAGS);
+	    count++;
 
 	}
 	else
 	    mp = mp->flink;
     }
 
-    if (rmp) {
+    if (had_start && (rmp != 0)) {
 	/* unget the unmatched suffix */
-	while (count > 0)
-	    itb_append(outp, unmatched[--count]);
+	while (suffix && (count > 0))
+	    (void)itb_append(outp, itb_values(unmatched)[--count]);
 	/* unget the mapping and elide correct number of recorded chars */
 	if (rmp->srv) {
-	    int remapflag;
+	    UINT remapflag;
 	    char *cp;
 	    /* cp = rmp->srv + cnt; */
 	    for (cp = rmp->srv; *cp; cp++)
-	    	;
+		;
 	    if (rmp->flags & MAPF_NOREMAP)
 		remapflag = NOREMAP;
 	    else
 		remapflag = 0;
 	    while (cp > rmp->srv)
-		itb_append(outp, char2int(*--cp)|remapflag);
+		(void)itb_append(outp, char2int(*--cp)|remapflag);
 	}
 	else {
-	    itb_append(outp, rmp->irv);
+	    (void)itb_append(outp, rmp->irv);
 	}
-	return matchedcnt;
     }
     else {	/* didn't find a match */
 	while (count > 0)
-	    itb_append(outp, unmatched[--count]);
-	return 0;
+	    (void)itb_append(outp, itb_values(unmatched)[--count]);
+	matchedcnt = 0;
     }
+    itb_free(&unmatched);
+    return matchedcnt;
 }
 
 #if NO_LEAKS
-static	void	free_maprec P((struct maprec **));
-
 static void
-free_maprec(p)
-struct maprec **p;
+free_maprec(struct maprec **p)
 {
 	struct	maprec *q;
 	if ((q = *p) != 0) {
@@ -852,7 +971,7 @@ struct maprec **p;
 }
 
 void
-map_leaks()
+map_leaks(void)
 {
 	free_maprec(&map_command);
 	free_maprec(&map_insert);
